"""Extracts the distributions associated with TMD module."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import tmd


def persistent_homology_angles(
    pop, threshold=2, neurite_type="basal_dendrite", feature="radial_distances"
):
    """Add the persistent homology extracted from a population of apicals to the distr dictionary.

    Each tree in the population is associated with a persistence barcode (diagram)
    and a set of angles that will be used as input for synthesis.

    Args:
        pop (neurom.core.population.Population): The given population.
        threshold (int): The minimum number of terminations.
        neurite_type (neurom.core.types.NeuriteType): Consider only the neurites of this type.
        feature (str): Use the specified TMD feature.
    """
    ph_ang = [
        tmd.methods.get_ph_angles(tree, feature=feature) for tree in getattr(pop, neurite_type)
    ]

    # Keep only the trees whose number of terminations is above the threshold
    # Saves the list of persistence diagrams for the selected neurite_type
    phs = [p for p in ph_ang if len(p) > threshold]
    min_bar_length = min(min(tmd.analysis.get_lengths(ph)) for ph in phs)

    return {"persistence_diagram": phs, "min_bar_length": min_bar_length}



"""Test neurots.generate.section code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import json
import os

import numpy as np
from numpy.testing import assert_array_almost_equal
from numpy.testing import assert_equal

from neurots.generate.algorithms.basicgrower import TrunkAlgo
from neurots.generate.algorithms.common import TMDStop
from neurots.generate.algorithms.tmdgrower import TMDAlgo
from neurots.generate.algorithms.tmdgrower import TMDApicalAlgo
from neurots.generate.algorithms.tmdgrower import TMDGradientAlgo
from neurots.generate.section import SectionGrowerPath
from neurots.generate.section import SectionGrowerTMD
from neurots.generate.tree import SectionParameters
from neurots.morphmath import sample

_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def _setup_test(Algo, Grower, custom_parameters=None):
    with open(os.path.join(_PATH, "dummy_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)["basal_dendrite"]

    with open(os.path.join(_PATH, "dummy_params.json"), encoding="utf-8") as f:
        parameters = json.load(f)["basal_dendrite"]
    parameters["bias_length"] = 0.5
    parameters["bias"] = 0.5
    parameters["has_apical_tuft"] = True

    if custom_parameters is not None:
        parameters.update(custom_parameters)

    np.random.seed(42)
    algo = Algo(distributions, parameters, [0, 0, 1])
    seg_len = sample.Distr(parameters["step_size"])

    section_parameters = SectionParameters(
        randomness=0.2, targeting=0.3, history=1.0 - 0.2 - 0.3, scale_prob=1.0
    )

    grower = Grower(
        parent=None,
        children=None,
        first_point=[1.1, 0.0, 0.0],
        direction=[0.57735, 0.57735, 0.57735],
        parameters=section_parameters,
        process="major",
        stop_criteria={"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)},
        step_size_distribution=seg_len,
        pathlength=0.0,
    )

    return algo, grower


def _assert_dict_or_array(dict1, dict2):
    assert_equal(set(dict1.keys()), set(dict2.keys()))
    for key in dict1.keys():
        if isinstance(dict1[key], np.ndarray):
            assert_array_almost_equal(dict1[key], dict2[key])
        elif isinstance(dict1[key], dict):
            _assert_dict_or_array(dict1[key], dict2[key])
        else:
            assert_equal(dict1[key], dict2[key], f"Error for key: {key}")


def test_TrunkAlgo():
    np.random.seed(0)
    custom_parameters = {
        "num_seg": 10,
        "branching_method": "random",
    }
    algo, grower = _setup_test(TrunkAlgo, SectionGrowerPath, custom_parameters)

    stop, num_sec = algo.initialize()
    assert stop == {"num_seg": 10}
    assert_equal(num_sec, 1)

    s1, s2 = algo.bifurcate(grower)
    _assert_dict_or_array(
        s1,
        {
            "direction": [-0.30524033, 0.30700944, 0.90142861],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": [-0.11067468, -0.97407245, 0.19731697],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)},
        },
    )


def test_TMDAlgo():
    algo, grower = _setup_test(TMDAlgo, SectionGrowerPath)

    stop, num_sec = algo.initialize()
    assert stop == {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)}
    assert_equal(num_sec, 10)
    assert_equal(grower.get_val(), 0)

    s1, s2 = algo.bifurcate(grower)
    _assert_dict_or_array(
        s1,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=1, term=124.8796, ref=0.0)},
        },
    )

    algo, grower = _setup_test(TMDAlgo, SectionGrowerTMD)

    stop, num_sec = algo.initialize()
    assert stop == {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)}
    assert_equal(num_sec, 10)
    assert_equal(grower.get_val(), 1.1)

    s1, s2 = algo.bifurcate(grower)
    _assert_dict_or_array(
        s1,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=1, term=124.8796, ref=0.0)},
        },
    )


def test_TMDAlgo_modify():
    def tmd_scale(barcode, thickness):
        # only the two first points of each bar are modified
        # because they correspond to spatial dimensions
        scaling_factor = np.ones(len(barcode[0]), dtype=float)
        scaling_factor[:2] = thickness
        return np.multiply(barcode, scaling_factor).tolist()

    def modify_barcode(ph, context, thickness=1150.0, thickness_reference=1150.0):
        # pylint: disable=unused-argument
        max_p = np.max(ph)
        scaling_reference = 1.0
        if 1 - max_p / thickness_reference < 0:  # If cell is larger than the layers
            scaling_reference = thickness_reference / max_p
        return tmd_scale(ph, scaling_reference * thickness / thickness_reference)

    custom_parameters = {
        "modify": {
            "funct": modify_barcode,
            "kwargs": {"thickness": 100, "thickness_reference": 1000},
        }
    }

    algo, grower = _setup_test(TMDAlgo, SectionGrowerPath, custom_parameters)

    stop, num_sec = algo.initialize()
    assert stop == {"TMD": TMDStop(bif_id=1, bif=0.9775, term_id=0, term=15.9798, ref=0.0)}
    assert_equal(num_sec, 10)

    s1, s2 = algo.bifurcate(grower)
    _assert_dict_or_array(
        s1,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=1.8525, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": [0.0, 0.0, 0.0],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=1.8525, term_id=1, term=12.488, ref=0.0)},
        },
    )


def test_TMDApicalAlgo():
    algo, grower = _setup_test(TMDApicalAlgo, SectionGrowerPath)

    stop, num_sec = algo.initialize()
    expected_stop = {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)}
    assert stop == expected_stop
    assert_equal(num_sec, 10)

    grower.id = 1
    s1, s2 = algo.bifurcate(grower)
    assert_equal(algo.apical_section, 1)
    _assert_dict_or_array(
        s1,
        {
            "direction": [0.57735, 0.57735, 0.57735],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": np.array([0.20348076, 0.54109933, 0.81597003]),
            "process": "secondary",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=1, term=124.8796, ref=0.0)},
        },
    )

    grower.stop_criteria["TMD"].bif_id = 2
    grower.stop_criteria["TMD"].bif = 18.5246
    grower.points[-1] *= 2
    grower.id = 2
    algo.bifurcate(grower)
    assert_equal(algo.apical_section, 2)

    # Find last bifurcation in secondary branch
    algo, grower = _setup_test(TMDApicalAlgo, SectionGrowerPath)
    grower.pathlength = 9999
    grower.process = "secondary"
    stop, num_sec = algo.initialize()
    grower.id = 3
    s1, s2 = algo.bifurcate(grower)
    assert_equal(algo.apical_section, 3)

    grower.stop_criteria["TMD"].bif_id = 2
    grower.stop_criteria["TMD"].bif = 18.5246
    grower.points[-1] *= 2
    grower.id = 4
    algo.bifurcate(grower)
    # This time the apical section was not updated
    assert_equal(algo.apical_section, 3)


def test_TMDGradientAlgo():
    algo, grower = _setup_test(TMDGradientAlgo, SectionGrowerPath)

    stop, num_sec = algo.initialize()
    expected_stop = {"TMD": TMDStop(bif_id=1, bif=9.7747, term_id=0, term=159.798, ref=0.0)}
    assert stop == expected_stop
    assert_equal(num_sec, 10)

    s1, s2 = algo.bifurcate(grower)
    _assert_dict_or_array(
        s1,
        {
            "direction": [0.57735, 0.57735, 0.57735],
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=0, term=159.798, ref=0.0)},
        },
    )

    _assert_dict_or_array(
        s2,
        {
            "direction": np.array([0.400454, 0.573604, 0.714573]),
            "process": "major",
            "first_point": [1.1, 0.0, 0.0],
            "stop": {"TMD": TMDStop(bif_id=2, bif=18.5246, term_id=1, term=124.8796, ref=0.0)},
        },
    )



"""Test neurots.astrocyte.tmd_utils code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from numpy import testing as npt

from neurots.astrocyte import tmd_utils as _tu


def _barcode():
    return [[230.75389593157917, 174.56457422613303], [1072.5500821184357, 0]]


def _barcode_list():
    return [
        [[230.75389593157917, 174.56457422613303], [1072.5500821184357, 0]],
        [[461.50779186315833, 349.12914845226607], [2145.1001642368715, 0]],
        [[692.2616877947376, 523.6937226783991], [3217.650246355307, 0]],
        [[923.0155837263167, 698.2582969045321], [4290.200328473743, 0]],
        [[1153.7694796578958, 872.8228711306651], [5362.750410592179, 0]],
    ]


def test_scale_barcode():

    barcode = _barcode()

    target_distance = 2.0 * 1072.5500821184357

    ph = _tu.scale_barcode(barcode, target_distance)

    expected = np.asarray(barcode) * 2.0
    npt.assert_allclose(ph, expected)

    target_distance = 0.8 * 1072.5500821184357

    ph = _tu.scale_barcode(barcode, target_distance)
    npt.assert_allclose(ph, barcode)


def test_barcodes_greater_than_distance():

    barcode_list = _barcode_list()

    result = _tu.barcodes_greater_than_distance(barcode_list, 3000.0)

    npt.assert_equal(len(result), 3)
    npt.assert_allclose(result, barcode_list[2:])



"""Models to create diameters of synthesized cells."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import inspect

import numpy as np
from morphio import SectionType


def section_filter(neuron, tree_type=None):
    """Filter all sections according to type."""
    if tree_type is None:
        return list(neuron.iter())
    return [i for i in neuron.iter() if i.type == tree_type]


def root_section_filter(neuron, tree_type=None):
    """Filter root sections according to type."""
    if tree_type is None:
        return list(neuron.root_sections)
    return [i for i in neuron.root_sections if i.type == tree_type]


def sample(data, random_generator=np.random):
    """Returns a value according to the input data.

    Args:
        data (list): The data to sample from.
        random_generator (numpy.random.Generator): The random number generator to use.
    """
    return random_generator.choice(data)


def section_lengths(section):
    """Computes all segment lengths within section.

    Args:
        section (morphio.Section): The section whose length is computed.

    Returns:
        float: The total length of the section.
    """
    return np.linalg.norm(section.points[1:] - section.points[:-1], axis=1)


def redefine_diameter_section(section, diam_ind, diam_new):
    """Hack to replace one diameter at index diam_ind with value diam_new.

    Args:
        section (morphio.mut.Section): The section whose diameter is updated.
        diam_ind (int): The index to replace.
        diam_new (float): The new diameter value.
    """
    section.diameters = np.array(
        section.diameters.tolist()[:diam_ind]
        + [diam_new]
        + section.diameters.tolist()[(diam_ind + 1) :]
    )
    if len(section.points) != len(section.diameters):
        raise Exception("Mismatch in dimensions of diameters.")


def bifurcator(initial_diam, num_children, rall_ratio, siblings_ratio):
    """Returns the computed bifurcation diameter.

    Args:
        initial_diam (float): The initial diameter value.
        num_children (int): The number of children.
        rall_ratio (float): The rall ratio.
        siblings_ratio (float): The sibling ratio.
    """
    # pylint: disable=assignment-from-no-return
    reduction_factor = np.power(
        1.0 + (num_children - 1) * np.power(siblings_ratio, rall_ratio),
        1.0 / rall_ratio,
    )
    return initial_diam / reduction_factor


def merger(section, trunk_diam, rall_ratio):
    """Returns the computed bifurcation diameter.

    Args:
        section (morphio.mut.Section): The section whose diameter is updated.
        trunk_diam (float): The trunk diameter.
        rall_ratio (float): The rall ratio.
    """
    # pylint: disable=assignment-from-no-return
    # diameters[0] is the duplicate point
    diameters_children = [ch.diameters[1] for ch in section.children]
    parent_d = np.power(
        np.sum([np.power(d, rall_ratio) for d in diameters_children]), 1.0 / rall_ratio
    )

    diam_val = parent_d if parent_d <= trunk_diam else np.max(diameters_children)
    redefine_diameter_section(section, len(section.diameters) - 1, diam_val)


def taper_section_diam_from_root(section, initial_diam, taper, min_diam=0.07, max_diam=100.0):
    """Corrects the diameters of a section.

    Args:
        section (morphio.mut.Section): The section whose diameters are updated.
        initial_diam (float): The initial diameter value.
        taper (float): The taper value.
        min_diam (float): The min diameter value.
        max_diam (float): The max diameter value.
    """
    diams = [initial_diam]

    if section.is_root:
        range_ = range(1, len(section.diameters))
    else:
        range_ = range(0, len(section.diameters) - 1)

    # lengths of each segments will be used for scaling of tapering
    lengths = [0] + section_lengths(section).tolist()
    taps = taper / np.sum(lengths)

    for i in range_:
        # Diameters decrease from root towards the tip
        new_diam = diams[-1] - taps * lengths[i]

        if new_diam >= max_diam:
            diams.append(max_diam)
        elif new_diam <= min_diam:
            diams.append(min_diam)
        else:
            diams.append(new_diam)

    section.diameters = np.array(diams, dtype=np.float32)


def taper_section_diam_from_tips(section, final_diam, taper, min_diam=0.07, max_diam=100.0):
    """Corrects the diameters of a section.

    Args:
        section (morphio.mut.Section): The section whose diameters are updated.
        final_diam (float): The final diameter value.
        taper (float): The taper value.
        min_diam (float): The min diameter value.
        max_diam (float): The max diameter value.
    """
    diams = [final_diam]
    # lengths of each segments will be used for scaling of tapering
    lengths = section_lengths(section)
    taps = taper / np.sum(lengths)

    for i in range(len(section.diameters) - 1):
        # Diameters increase from tip towards the root
        new_diam = diams[-1] + taps * lengths[len(section.diameters) - 2 - i]

        if new_diam >= max_diam:
            diams.append(max_diam)
        elif new_diam <= min_diam:
            diams.append(min_diam)
        else:
            diams.append(new_diam)

    diams.reverse()
    section.diameters = np.array(diams, dtype=np.float32)


def smooth_section_diam(section, min_diam=0.07):
    """Corrects the diameters of a section by smoothing between initial and final diameters.

    Args:
        section (morphio.mut.Section): The section whose diameters are updated.
        min_diam (float): The min diameter value.
    """

    def sec_mean_taper(sec):
        """Returns the mean tapering of section."""
        min_diam = min(sec.diameters)
        lengths = [np.linalg.norm(i) for i in section.points[1:] - section.points[:-1]]
        di_li = np.sum(
            [
                (sec.diameters[i] + sec.diameters[i + 1]) / 2.0 * lengths[i]
                for i in range(len(sec.points) - 1)
            ]
        )
        return (di_li - min_diam * np.sum(lengths)) / np.sum(lengths)

    taper = sec_mean_taper(section)
    taper_section_diam_from_root(
        section,
        section.diameters[0],
        taper,
        min_diam=min_diam,
        max_diam=np.max(section.diameters),
    )


def diametrize_from_root(
    neuron,
    neurite_type=None,
    *,
    model_params,
    random_generator=np.random,
):  # pylint: disable=too-many-locals
    """Corrects the diameters of a morphio-neuron according to the model.

    Starts from the root and moves towards the tips.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        neurite_type (morphio.SectionType): Only the neurites of this type are diametrized.
        model_params (dict): The model parameters.
        random_generator (numpy.random.Generator): The random number generator to use.
    """
    for r in root_section_filter(neuron, tree_type=neurite_type):

        model = model_params[r.type.name]  # Selected by the root type.
        trunk_diam = sample(model["trunk"], random_generator)
        min_diam = np.min(model["term"])
        status = {s.id: False for s in r.iter()}

        active = [r]

        while active:
            for section in list(active):

                if section.is_root:
                    taper = sample(model["trunk_taper"], random_generator)
                    init_diam = trunk_diam
                else:
                    taper = sample(model["taper"], random_generator)
                    init_diam = section.diameters[0]

                taper_section_diam_from_root(
                    section,
                    init_diam,
                    taper=taper,
                    min_diam=min_diam,
                    max_diam=trunk_diam,
                )
                status[section.id] = True  # Tapering of section complete.
                active.remove(section)

                children = np.array(section.children)

                if len(children) > 1:
                    d1 = bifurcator(
                        section.diameters[-1],
                        len(children),
                        rall_ratio=model["Rall_ratio"],
                        siblings_ratio=model["siblings_ratio"],
                    )

                    for i, ch in enumerate(children):
                        new_diam = d1 if i == 0 else d1 * model["siblings_ratio"]
                        redefine_diameter_section(ch, 0, new_diam)
                        active.append(ch)

        for section in r.iter():
            if not section.is_root:  # if section is not root replace diameter with parent
                # Ensures duplicate points consistency. First point will be removed while written.
                redefine_diameter_section(section, 0, section.parent.diameters[-1])


def diametrize_from_tips(neuron, neurite_type=None, *, model_params, random_generator=np.random):
    """Corrects the diameters of a morphio-neuron according to the model.

    Starts from the tips and moves towards the root.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        neurite_type (morphio.SectionType): Only the neurites of this type are diametrized.
        model_params (dict): The model parameters.
        random_generator (numpy.random.Generator): The random number generator to use.
    """
    for r in root_section_filter(neuron, tree_type=neurite_type):
        model = model_params[r.type.name]  # Selected by the root type.
        trunk_diam = sample(model["trunk"], random_generator)
        min_diam = np.min(model["term"])
        tips = [s for s in r.iter() if not s.children]
        status = {s.id: False for s in r.iter()}

        for tip in tips:
            redefine_diameter_section(
                tip, len(tip.diameters) - 1, sample(model["term"], random_generator)
            )

        active = tips

        while active:
            for section in list(active):
                taper = (
                    sample(model["trunk_taper"], random_generator)
                    if section.is_root
                    else sample(model["taper"], random_generator)
                )

                taper_section_diam_from_tips(
                    section,
                    section.diameters[-1],
                    taper=taper,
                    min_diam=min_diam,
                    max_diam=trunk_diam,
                )
                status[section.id] = True  # Tapering of section complete.
                active.remove(section)

                if not section.is_root:
                    par = section.parent
                    if np.alltrue([status[ch.id] for ch in par.children]):
                        # Assign a new diameter to the last point if section is not terminal
                        merger(par, trunk_diam, model["Rall_ratio"])
                        active.append(par)

        for section in r.iter():
            if not section.is_root:  # if section is not root replace diameter with parent
                # Ensures duplicate points consistency. First point will be removed while written.
                redefine_diameter_section(section, 0, section.parent.diameters[-1])


def diametrize_constant_per_section(neuron, neurite_type=None):
    """Corrects the diameters of a morphio-neuron to make them constant per section.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        neurite_type (morphio.SectionType): Only the neurites of this type are diametrized.
    """
    for sec in section_filter(neuron, neurite_type):
        mean_diam = np.mean(sec.diameters)
        sec.diameters = mean_diam * np.ones(len(sec.diameters))


def diametrize_constant_per_neurite(neuron, neurite_type=None):
    """Corrects the diameters of a morphio-neuron to make them constant per neurite.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        neurite_type (morphio.SectionType): Only the neurites of this type are diametrized.
    """
    roots = root_section_filter(neuron, neurite_type)

    for root in roots:
        mean_diam = np.mean(np.hstack([sec.diameters for sec in root.iter()]))
        for sec in root.iter():
            sec.diameters = mean_diam * np.ones(len(sec.diameters))


def diametrize_smoothing(neuron, neurite_type=None):
    """Corrects the diameters of a morphio-neuron, by smoothing them within each section.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        neurite_type (morphio.SectionType): Only the neurites of this type are diametrized.
    """
    for sec in section_filter(neuron, neurite_type):
        smooth_section_diam(sec)


diam_methods = {
    "M1": diametrize_constant_per_neurite,
    "M2": diametrize_constant_per_section,
    "M3": diametrize_smoothing,
    "M4": diametrize_from_root,
    "M5": diametrize_from_tips,
}


def build(
    neuron,
    input_model=None,
    neurite_types=None,
    diam_method=None,
    diam_params=None,
    random_generator=np.random,
):
    """Diametrize according to the selected method.

    Args:
        neuron (morphio.mut.Morphology): The morphology that will be diametrized.
        input_model (dict): The model parameters.
        neurite_types (list[str]): Only the neurites of these types are diametrized.
        diam_method (str or callable): The name of the diametrization method.
        diam_params (dict): The parameters passed to the diametrization method.
        random_generator (numpy.random.Generator): The random number generator to use.

    If ``diam_method`` is a string matching the models in ``diam_methods`` it will use an internal
    diametrizer. If a function is provided, it will use this function to diametrize the cells. This
    function should have the following arguments:

    * **neuron** (*morphio.mut.Morphology*): The morphology that will be diametrized.
    * **tree_type** (*str*): Only the neurites of this type are diametrized.
    * **model_params** (*dict*): The model parameters (optional).
    * **diam_params** (*dict*): The parameters passed to the diametrization method (optional).
    * **random_generator** (*numpy.random.Generator*): The random number generator to use
        (optional).

    and should only update the neuron object.
    """
    if neurite_types is None:
        neurite_types = [SectionType.apical_dendrite, SectionType.basal_dendrite]

    if isinstance(diam_method, str):
        try:
            diam_method = diam_methods[diam_method]
        except KeyError as exc:
            raise KeyError(
                "The name of the diametrization method is unknown: "
                f"'{diam_method}' is not in {list(diam_methods.keys())}"
            ) from exc

    elif not hasattr(diam_method, "__call__"):
        raise ValueError(f"Diameter method not understood, we got {diam_method}")

    for tree_type in neurite_types:
        if isinstance(tree_type, str):
            tree_type = getattr(SectionType, tree_type)
        param_signature = list(inspect.signature(diam_method).parameters.keys())
        optional_kw = {}
        if "model_params" in param_signature:
            optional_kw["model_params"] = input_model
        if "diam_params" in param_signature:
            optional_kw["diam_params"] = diam_params
        if "random_generator" in param_signature:
            optional_kw["random_generator"] = random_generator
        diam_method(neuron, tree_type, **optional_kw)



"""Test JSON schema validation."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import json
import os

from jsonschema import validate

_PATH = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_json_schema():

    with open(os.path.join(_PATH, "dummy_distribution.json"), encoding="utf-8") as f:
        data = json.load(f)

    validate(
        data,
        schema={
            "definitions": {
                "histogram": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "data": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {"bins": {"type": "array"}, "weights": {"type": "array"}},
                        }
                    },
                },
                "dendrite": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "trunk": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "orientation_deviation": {"$ref": "#/definitions/histogram"},
                                "absolute_elevation_deviation": {"$ref": "#/definitions/histogram"},
                                "azimuth": {"$ref": "#/definitions/uniform_distrib"},
                            },
                        },
                        "num_trees": {"type": "object"},
                        "min_bar_length": {"type": "number"},
                        "persistence_diagram": {"type": "array"},
                        "filtration_metric": {"type": "string"},
                    },
                },
                "norm_distrib": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "norm": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "mean": {"type": "number"},
                                "std": {"type": "number"},
                            },
                        }
                    },
                },
                "uniform_distrib": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "uniform": {
                            "type": "object",
                            "additionalProperties": False,
                            "properties": {
                                "min": {"type": "number"},
                                "max": {"type": "number"},
                            },
                        }
                    },
                },
                "diameter": {
                    "type": "object",
                    "additionalProperties": True,
                },
            },
            "type": "object",
            "properties": {
                "soma": {
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {"size": {"$ref": "#/definitions/norm_distrib"}},
                },
                "basal_dendrite": {"$ref": "#/definitions/dendrite"},
                "apical_dendrite": {"$ref": "#/definitions/dendrite"},
                "axon": {"$ref": "#/definitions/dendrite"},
                "diameter": {"$ref": "#/definitions/diameter"},
            },
            "required": ["soma", "axon", "basal_dendrite", "apical_dendrite", "diameter"],
            "additionalProperties": False,
        },
    )



"""NeuroTS utils used by multiple tools."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import warnings
from copy import deepcopy

import numpy as np


class NeuroTSError(Exception):
    """Raises NeuroTS error."""


def format_values(obj, decimals=None):
    """Format values of an object recursively."""
    if isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, np.floating):
        obj = float(obj)
        if decimals is not None:
            obj = round(obj, ndigits=decimals)
    elif isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, (np.bool8, np.bool_)):
        obj = bool(obj)
    elif isinstance(obj, dict):
        obj = {k: format_values(v, decimals=decimals) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple, set)):
        obj = type(obj)([format_values(i, decimals=decimals) for i in obj])
    return obj


def neurite_type_warning(key):
    """Print a deprecation warning for old neurite_type key."""
    warnings.warn(
        f"The '{key}' property is deprecated, please use '{key}_dendrite' instead",
        DeprecationWarning,
    )


def convert_from_legacy_neurite_type(data):
    """Convert legacy neurite type names, basal -> basal_dendrite and apical -> apical_dendrite."""
    old_data = deepcopy(data)
    for key, _data in old_data.items():

        if key == "apical":
            neurite_type_warning(key)
            data["apical_dendrite"] = data.pop("apical")
            key = "apical_dendrite"
        if key == "basal":
            neurite_type_warning(key)
            data["basal_dendrite"] = data.pop("basal")
            key = "basal_dendrite"

        if isinstance(_data, dict):
            data[key] = convert_from_legacy_neurite_type(data[key])

        if isinstance(_data, list):
            for i, d in enumerate(_data):
                if d == "apical":
                    neurite_type_warning(key)
                    data[key][i] = "apical_dendrite"
                if d == "basal":
                    neurite_type_warning(key)
                    data[key][i] = "basal_dendrite"

    return data



"""NeuroTS class: Soma."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np
from scipy.spatial import ConvexHull

try:
    # The QhulError was moved in scipy >= 1.8 so if the import fails the old location is imported
    from scipy.spatial import QhullError
except ImportError:  # pragma: no cover
    from scipy.spatial.qhull import QhullError

from neurots.generate import orientations
from neurots.morphmath import rotation
from neurots.utils import NeuroTSError

L = logging.getLogger()


class Soma:
    """NeuroTS soma data structure.

    It contains the soma points and radius.

    Args:
        center (numpy.ndarray): The center of the soma.
        radius (float): The radius of the soma.
        points (list[list[float]], optional): An iterable of 3D points.
    """

    def __init__(self, center, radius, points=None):
        self.radius = float(radius)
        self._center = np.asarray(center, dtype=np.float64)
        self.points = [] if points is None else points

    @property
    def points(self):
        """Return the points list."""
        return self._points

    @points.setter
    def points(self, points):
        """Set the points list as a list of numpy arrays."""
        self._points = [np.asarray(p, dtype=np.float64) for p in points]

    @property
    def center(self):
        """Get the center."""
        return self._center

    @center.setter
    def center(self, xyz):
        """Set center as a numpy array."""
        self._center = np.asarray(xyz, dtype=np.float64)

    def point_from_trunk_direction(self, phi, theta):
        """Return the coordinates of a point on the soma surface from the given theta, phi angles.

        Args:
            phi (list[float]): The polar angle (i.e. on the z-axis).
            theta (list[float]): The azimuthal angle (i.e. on the x-y plane).
        """
        return np.array(
            [
                self.center[0] + self.radius * np.cos(phi) * np.sin(theta),
                self.center[1] + self.radius * np.sin(phi) * np.sin(theta),
                self.center[2] + self.radius * np.cos(theta),
            ],
            dtype=np.float64,
        )

    def orientation_from_point(self, point):
        """Return the orientation from the soma center to a point on the soma surface."""
        if np.allclose(point, self.center):
            raise ValueError("Point overlaps with soma center.")

        point = np.asarray([point])
        return orientations.points_to_orientations(self.center, point)[0]

    def contour_point(self, point):
        """Create a contour point from a given point.

        Keeps the x-y coordinates of the input point but replaces the third (z) coordinate with the
        equivalent soma-z in order to create a contour at the soma level.
        """
        return np.array((point[0], point[1], self.center[2]), dtype=np.float64)


class SomaGrower:
    """Soma class.

    Args:
        soma (Soma): The soma that will grow.
        context (Any): The context used for the section.
        rng (numpy.random.Generator): The random number generator to use.
    """

    def __init__(self, soma, context=None, rng=np.random):
        """Constructor of the Soma class."""
        self.soma = soma
        self.context = context  # for future, hypothetical use
        self._rng = rng

    def add_points_from_trunk_angles(self, trunk_angles, z_angles, phi_interval=None):
        """Generate points on the soma surface from a list of angles.

        Args:
            trunk_angles (list[float]): The polar angles (phi in spherical coordinates).
            z_angles (list[float]): The azimuthal angles (theta in spherical coordinates).
            phi_interval (tuple[float, float]): The interval in which the trunks should be added.

        Returns:
            list[list[float]]: The new points.
        """
        phis, thetas = orientations.trunk_to_spherical_angles(trunk_angles, z_angles, phi_interval)
        new_directions = orientations.spherical_angles_to_orientations(phis, thetas)
        return self.add_points_from_orientations(new_directions)

    def add_points_from_trunk_absolute_orientation(
        self, orientation, trunk_absolute_angles, z_angles
    ):
        """Generate points on the soma surface from a direction and a list of angles.

        Args:
            orientation (list[float]): The trunk orientation.
            trunk_absolute_angles (list[float]): The polar angles (phi in spherical coordinates).
            z_angles (list[float]): The azimuthal angles (theta in spherical coordinates).

        Returns:
            list[list[float]]: The new points.
        """
        phis, thetas = orientations.trunk_absolute_orientation_to_spherical_angles(
            orientation, trunk_absolute_angles, z_angles
        )
        new_directions = orientations.spherical_angles_to_orientations(phis, thetas)
        return self.add_points_from_orientations(new_directions)

    def add_points_from_orientations(self, vectors):
        """Generate points on the soma surface from a list of unit vectors.

        Args:
            vectors (list[list[float]]): The list of orientations.

        Returns:
            list[list[float]]: The new points.
        """
        new_points = []

        for vect in vectors:

            phi, theta = rotation.spherical_from_vector(vect)
            point = self.soma.point_from_trunk_direction(phi, theta)
            new_points.append(point)

        self.soma.points.extend(new_points)
        return new_points

    def interpolate(self, points, interpolation=10):
        """Return the interpolation points on the convex hull.

        Finds the convex hull from a list of points and returns a number of interpolation points
        that belong on this convex hull.

        Args:
            points (list[float]): Initial set of points.
            interpolation (int): The the minimum number of points to be generated.

        Returns:
            list[list[float]]: The list of interpolated points.
        """
        interpolation = np.max([3, interpolation])  # soma must have at least 3 points

        if len(points) >= interpolation:
            points_to_interpolate = points
        else:
            # Adds points from circle circumference to the soma points.
            angles = 2.0 * np.pi * self._rng.random(interpolation - len(points))
            x = self.soma.radius * np.sin(angles) + self.soma.center[0]
            y = self.soma.radius * np.cos(angles) + self.soma.center[1]
            z = np.full_like(angles, self.soma.center[2])
            points_to_interpolate = points + [[i, j, k] for i, j, k in zip(x, y, z)]

        # a convex hull from 2D points is guaranteed to be ordered
        xy_points = np.asarray(points_to_interpolate)[:, :2]

        try:
            selected = ConvexHull(xy_points).vertices
            return [
                points_to_interpolate[index]
                for index in selected  # pylint: disable=not-an-iterable
            ]
        except QhullError as e:
            raise NeuroTSError(
                "Warning! Convex hull failed, contour soma generation failed."
            ) from e

    def build(self, method="contour"):
        """Generates a soma, depending on the method used.

        The points will be saved into the neuron object and consist the first section of the cell.
        If interpolation is selected points will be generated until the expected number of soma
        points is reached.

        Args:
            method (str): The method used to build the soma.

        Returns:
            tuple[list[float], list[float]]: The points and diameters of the built soma.
        """
        if method == "contour":
            return self._contour_soma()
        elif method == "one_point":
            return self._one_point_soma()
        else:
            return self._original_soma()

    def _one_point_soma(self):
        """Generate a single point soma.

        This kind of soma is represented by a sphere including the center and the diameter.

        Returns:
            tuple[list[float], list[float]]: The points and diameters of the built soma.
        """
        soma_points = [self.soma.center]
        soma_diameters = [2.0 * self.soma.radius]
        return soma_points, soma_diameters

    def _contour_soma(self):
        """Generate a contour soma, that consists of all soma points.

        The contour must contain at least three points.

        Returns:
            tuple[list[float], list[float]]: The points and diameters of the built soma.
        """
        contour = [self.soma.contour_point(p) for p in self.soma.points]
        soma_pts = self.interpolate(contour)
        return soma_pts, np.zeros(len(soma_pts), dtype=np.float64)

    def _original_soma(self):
        """Returns the original soma points.

        Returns:
            tuple[list[float], list[float]]: The points and diameters of the built soma.
        """
        return self.soma.points, np.zeros(len(self.soma.points), dtype=np.float64)



"""Test tree generation."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
# pylint: disable=protected-access
import json
import os

import numpy as np
import pytest
from numpy import testing as npt

from neurots import NeuronGrower
from neurots.generate.tree import TreeGrower
from neurots.generate.tree import _create_section_parameters
from neurots.utils import NeuroTSError

_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def test_create_section_parameters__normal_input():

    input_dict = {"randomness": 0.3, "targeting": 0.6}

    parameters = _create_section_parameters(input_dict)

    npt.assert_almost_equal(parameters.randomness, 0.3)
    npt.assert_almost_equal(parameters.targeting, 0.6)
    npt.assert_almost_equal(parameters.history, 1.0 - 0.6 - 0.3)


def test_create_section_parameters__exceed_bounds():

    input_dict = {"randomness": 1.1, "targeting": -0.1}
    parameters = _create_section_parameters(input_dict)

    npt.assert_almost_equal(parameters.randomness, 1.0)
    npt.assert_almost_equal(parameters.targeting, 0.0)
    npt.assert_almost_equal(parameters.history, 0.0)

    input_dict = {"randomness": -0.1, "targeting": 1.0}
    parameters = _create_section_parameters(input_dict)

    npt.assert_almost_equal(parameters.randomness, 0.0)
    npt.assert_almost_equal(parameters.targeting, 1.0)
    npt.assert_almost_equal(parameters.history, 0.0)


def test_create_section_parameters__sum_to_one_error():

    input_dict = {"randomness": 2.0, "targeting": 2.0}

    with pytest.raises(NeuroTSError):
        _create_section_parameters(input_dict)


def test_TreeGrower():
    np.random.seed(0)
    with open(os.path.join(_path, "bio_distribution.json"), encoding="utf-8") as f:
        distributions = json.load(f)

    with open(os.path.join(_path, "bio_path_params.json"), encoding="utf-8") as f:
        params = json.load(f)

    grower = NeuronGrower(input_distributions=distributions, input_parameters=params)
    grower._grow_soma()

    # Test order_per_process()
    sections = [i.active_sections[0] for i in grower.active_neurites]
    for num, i in enumerate(sections):
        i.process = str(len(sections) - num - 1)
    res = TreeGrower.order_per_process(sections)
    for num, i in enumerate(res):
        assert i == sections[len(sections) - num - 1], (
            i,
            sections[len(sections) - num - 1],
        )



"""NeuroTS class: Section grower."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from collections import deque

import numpy as np
from numpy.linalg import norm as vectorial_norm  # vectorial_norm used for array of vectors

from neurots.morphmath.utils import get_random_point  # norm used for single vectors
from neurots.morphmath.utils import norm

MEMORY = 5
DISTANCE_MIN = 1e-8

# Memory decreases with distance from current point
WEIGHTS = np.exp(np.arange(1, MEMORY + 1) - MEMORY)


class SectionGrower:
    """Class for the section growth.

    A section is a list of points in 4D space (x, y, x, r) that are sequentially connected to each
    other. This process generates a tubular morphology that resembles a random walk.

    Args:
        parent (morphio.Section): The parent of the section.
        children (int): The number of children.
        first_point (list[float]): The first point of the section.
        direction (list[float]): The first point of the section.
        parameters (SectionParameters): The parameters used to grow the section.
        process (str): The process type.
        stop_criteria (dict): The stop criteria used for this section.
        step_size_distribution (Distr): The step size distribution.
        pathlength (float): The path length of the section.
        context (Any): The context used for the section.
        random_generator (numpy.random.Generator): The random number generator to use.
    """

    # pylint: disable-msg=too-many-arguments
    def __init__(
        self,
        parent,
        children,
        first_point,
        direction,
        parameters,
        process,
        stop_criteria,
        step_size_distribution,
        pathlength,
        context=None,
        random_generator=np.random,
    ):
        self.parent = parent
        self.id = None
        assert not np.isclose(vectorial_norm(direction), 0.0), "Nan direction not recognized"
        self.direction = direction / vectorial_norm(direction)
        self.children = children
        self.points = [np.array(first_point[:3])]

        self.params = parameters

        self.stop_criteria = stop_criteria
        self.process = process
        self.latest_directions = deque(maxlen=MEMORY)
        self.context = context
        self._rng = random_generator
        self.step_size_distribution = step_size_distribution
        self.pathlength = 0 if parent is None else pathlength

    @property
    def last_point(self):
        """Returns the last point of the section."""
        return self.points[-1]

    def update_pathlength(self, length):
        """Increases the path distance."""
        self.pathlength += length

    def next_point(self, current_point):
        """Returns the next point depending on the growth method and the previous point."""
        direction = (
            self.params.targeting * self.direction
            + self.params.randomness * get_random_point(random_generator=self._rng)
            + self.params.history * self.history()
        )

        direction = direction / vectorial_norm(direction)
        seg_length = self.step_size_distribution.draw_positive()
        next_point = current_point + seg_length * direction
        self.update_pathlength(seg_length)
        return next_point, direction

    def first_point(self):
        """Generates the first point of the section from the growth method and the previous point.

        This gives flexibility to implement a specific computation for the first point, and ensures
        the section has at least one point.

        .. warning::
            The growth process cannot terminate before this point, as a first point will always be
            added to an active section.
        """
        direction = self.params.targeting * self.direction + self.params.history * self.history()

        direction = direction / vectorial_norm(direction)
        seg_length = self.step_size_distribution.draw_positive()
        point = self.last_point + seg_length * direction
        self.update_pathlength(seg_length)

        self.latest_directions.append(direction)
        self.points.append(point)
        self.post_next_point()

    def check_stop(self):
        """Checks if any num_seg criteria is fulfilled.

        If it is, it returns False and the growth stops.
        """
        return len(self.points) < self.stop_criteria["num_seg"]

    def history(self):
        """Returns a combination of the sections history."""
        n_points = len(self.latest_directions)

        if n_points == 0:
            return np.zeros(3)

        history = np.dot(WEIGHTS[MEMORY - n_points :], self.latest_directions)

        distance = vectorial_norm(history)
        if distance > DISTANCE_MIN:
            history /= distance

        return history

    def next(self):
        """Creates one point and returns the next state: bifurcate, terminate or continue."""
        curr_point = self.last_point
        point, direction = self.next_point(curr_point)
        self.latest_directions.append(direction)
        self.points.append(point)
        self.post_next_point()

        if self.check_stop():
            return "continue"

        if self.children == 0:
            return "terminate"

        return "bifurcate"

    def post_next_point(self):
        """A method to perform actions after `self.next_point()` has been called."""


class SectionGrowerExponentialProba(SectionGrower):
    """Abstract class for exponentially decreasing bifurcation and termination probabilities.

    The parameter lambda defines the slope of the exponential.
    The parameter that follows the exponential must be defined in the derived class.
    """

    def _check(self, value, which):
        crit = getattr(self.stop_criteria["TMD"], which)
        scale_prob = self.params.scale_prob
        assert scale_prob > 0
        x = crit - value
        if x < 0:
            # no need to exponentiate, the comparison below automatically resolves to `True`
            return True
        # Check if close enough to exp( distance * scale_prob)
        return self._rng.random() < np.exp(-x * scale_prob)

    def check_stop(self):
        """Probabilities of bifurcating and stopping are proportional `exp(-distance * lambda)`."""
        if len(self.points) < 2:
            return True

        val = self.get_val()

        if self._check(val, "bif"):
            self.children = 2.0
            return False

        if self._check(val, "term"):
            self.children = 0.0
            return False

        return True

    def get_val(self):
        """Placeholder for any function."""
        raise NotImplementedError("Attempt to use abstract class")


class SectionGrowerTMD(SectionGrowerExponentialProba):
    """Class for the TMD section growth."""

    def get_val(self):
        """Returns radial distance."""
        return norm(np.subtract(self.last_point, self.stop_criteria["TMD"].ref))


class SectionGrowerPath(SectionGrowerExponentialProba):
    """Class for the TMD path based section growth."""

    def get_val(self):
        """Returns path distance."""
        return self.pathlength



"""Example for comparing two populations using
   the violin plots from their morphometrics"""
import neurom as nm
import numpy as np
import pandas
import pylab as plt
import seaborn

feat_list = [
    "number_of_neurites",
    "number_of_sections_per_neurite",
    "number_of_leaves",
    "number_of_bifurcations",
    "section_lengths",
    "section_tortuosity",
    "section_radial_distances",
    "section_path_distances",
    "section_branch_orders",
    "remote_bifurcation_angles",
]


feat_names = [
    "Number of neurites",
    "Number of sections",
    "Number of terminations",
    "Number of bifurcations",
    "Section lengths",
    "Section tortuosity",
    "Section radial distances",
    "Section path distances",
    "Section branch orders",
    "Remote bif angles",
]


def get_features_medians(object1, object2, flist=feat_list, neurite_type=nm.BASAL_DENDRITE):
    """Computes features from module mod"""
    collect_all = []

    for feat in flist:

        feature_pop = [np.median(nm.get(feat, obj, neurite_type=neurite_type)) for obj in object1]
        feature_neu = [np.median(nm.get(feat, obj, neurite_type=neurite_type)) for obj in object2]

        collect_all.append([feature_pop, feature_neu])

    return collect_all


def get_features_all(object1, object2, flist=feat_list, neurite_type=nm.BASAL_DENDRITE):
    """Computes features from module mod"""
    collect_all = []

    for feat in flist:
        feature_pop = []
        for obj in object1:
            feature_pop = feature_pop + nm.get(feat, obj, neurite_type=neurite_type).tolist()
        feature_neu = []
        for obj in object2:
            feature_neu = feature_neu + nm.get(feat, obj, neurite_type=neurite_type).tolist()

        collect_all.append([feature_pop, feature_neu])

    return collect_all


def transform2DataFrame(data, pop_names, flist=feat_names):
    """Returns a DataFrame in the appropriate
    format from a set of features"""
    values = []
    names = []
    feat = []
    # Loop through each feature
    for i, d1 in enumerate(data):
        m = np.mean(d1[0])
        st = np.std(d1[0])
        # Loop through populations
        for j, d2 in enumerate(d1):
            values = values + [(d3 - m) / st for d3 in d2]
            names = names + len(d2) * [pop_names[j]]
            feat = feat + len(d2) * [flist[i]]

    frame = pandas.DataFrame({"Data": names, "Values": values, "Morphological features": feat})

    return frame


def plot_violins(data, x="Morphological features", y="Values", hues="Data", **kwargs):
    """Plots the split violins of all features"""
    axs = seaborn.violinplot(x=x, y=y, hue=hues, data=data, palette="muted", split=True, **kwargs)
    plt.xticks(rotation=20)
    plt.tight_layout(True)
    return axs



"""Test neurots.morphmath.bifurcation code."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# pylint: disable=missing-function-docstring
import numpy as np
from numpy.testing import assert_array_almost_equal

from neurots.morphmath import bifurcation as _bf


def test_get_bif_directional():
    assert_array_almost_equal(
        _bf.directional([0.0, 1.0, 0.0], [0.0, 0.0, 0.0, np.pi / 2]),
        (np.array([0.0, 1.0, 0.0]), np.array([0.0, 0.0, 1.0])),
    )
    assert_array_almost_equal(
        _bf.directional([0.0, 1.0, 0.0], [0.0, 0.0, 0.0, 0]),
        (np.array([0.0, 1.0, 0.0]), np.array([0.0, 1.0, 0.0])),
    )
    assert_array_almost_equal(
        _bf.directional([0.0, 1.0, 0.0], [0.0, 0.0, 0.0, np.pi]),
        (np.array([0.0, 1.0, 0.0]), np.array([0.0, -1.0, 0.0])),
    )
    assert_array_almost_equal(
        _bf.directional([0.0, 1.0, 0.0], [0.0, 0.0, 0.0, np.pi / 4]),
        (
            np.array([0.0, 1.0, 0.0]),
            np.array([0.0, 1.0 / np.sqrt(2), 1.0 / np.sqrt(2)]),
        ),
    )


def test_get_bif_bio_oriented():
    assert_array_almost_equal(
        _bf.bio_oriented([0.0, 1.0, 0.0], [0.0, 0.0, 0.0, np.pi / 4]),
        (
            np.array([0.0, 1.0, 0.0]),
            np.array([0.0, 1.0 / np.sqrt(2), 1.0 / np.sqrt(2)]),
        ),
    )
    assert_array_almost_equal(
        _bf.bio_oriented([0.0, 1.0, 0.0], [0.0, np.pi / 4, 0.0, np.pi / 4]),
        (np.array([0.0, 1.0 / np.sqrt(2), 1.0 / np.sqrt(2)]), np.array([0.0, 0, 1])),
    )


def test_get_bif_symmetric():
    assert_array_almost_equal(
        _bf.symmetric([0, 0, 1], [1, 1, 1, 1]),
        [[0.0, -0.479426, 0.877583], [0.0, 0.479426, 0.877583]],
    )



"""Functionality used by multiple algorithms."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np

from neurots.morphmath import bifurcation as _bif

bif_methods = {
    "bio_oriented": _bif.bio_oriented,
    "symmetric": _bif.symmetric,
    "directional": _bif.directional,
    "random": _bif.random,
}


def checks_bif_term(ref, bif, term, target_length):
    """Check bif/term.

    Args:
        bif (float): The bifurcation value.
        term (float): The termination value.
        target_length (float): The target value.

    Returns:
        bool: ``True`` if:

        1. ``Ref < Bif < Term`` unless Bif is infinite, then ``Ref < Term``.
        2. ``target_length >= (Bif - Ref)``.
        3. ``target_length >= (Term - Ref)``.

        Otherwise returns ``False``.
    """
    term_cond = 0 < term - ref <= target_length

    if np.isinf(bif):
        return term_cond

    bif_cond = 0 < bif - ref <= target_length

    return term_cond and bif_cond and term > bif


def section_data(direction, first_point, stop_criteria, process_type):
    """Generate section data dictionary from arguments."""
    return {
        "direction": direction,
        "first_point": first_point,
        "stop": stop_criteria,
        "process": process_type,
    }


class TMDStop:
    """Class to define the data for stop criteria based on the TMD method.

    Args:
        bif_id (int): The bifurcation ID.
        bif (float): The bifurcation value.
        term_id (int): The termination ID.
        term (float): The termination value.
        ref (float): The reference value (i.e for path or radial distances).
    """

    def __init__(self, bif_id, bif, term_id, term, ref):
        """Initialization of TMDStop class with parameters."""
        self.bif_id = bif_id
        self.bif = bif
        self.term_id = term_id
        self.term = term
        self.ref = ref
        # self.verify()

    def __str__(self):
        return (
            f"(Ref: {self.ref}, BifID: {self.bif_id}, Bif: {self.bif}, TermID: {self.term_id}, "
            f"Term: {self.term})"
        )

    def printme(self):
        """Print all features."""
        print(self)  # pragma: no cover

    def verify(self):
        """Check the validity of the TMDStop.

        Validity is defined by ``bif < term`` unless bif in infinity.
        Also both ``bif`` and ``term`` have to be larger than the reference distance.

        Returns:
            bool: ``True`` if stop is valid, ``False`` otherwise.
        """
        # If no bifurcation the bif comparison is meaningless
        if np.isinf(self.bif) and (self.ref <= self.term):
            return True
        if self.ref <= self.bif <= self.term:
            return True
        return False

    def update_bif(self, bif_id, bif):
        """Set new values to bifurcation."""
        self.bif_id = bif_id
        self.bif = bif

    def update_term(self, term_id, term):
        """Set new values to termination."""
        self.term_id = term_id
        self.term = term

    def child_length(self):
        """Return the child length.

        Return the absolute difference between bifurcation and termination, which defines the
        length of the bar.
        """
        return abs(self.term - self.bif)

    def expected_bifurcation_length(self):
        """Compute an estimate for the length of the branch if a bifurcation occurs.

        That will happen at distance ``bif`` and therefore the expected length is:
        ``(bifurcation - current length)``.
        If ``bifurcation`` is ``inf``, the expected length is zero.
        """
        if np.isinf(self.bif):
            return 0.0
        return abs(self.bif - self.ref)

    def expected_termination_length(self):
        """Compute an estimate for the length of the branch if a termination occurs.

        That will happen at distance ``term`` and therefore the expected length is:
        ``(termination - current length)``.
        If ``termination`` is ``inf``, the expected length is zero.
        """
        if np.isinf(self.term):
            return 0.0
        return abs(self.term - self.ref)

    def expected_maximum_length(self):
        """Return the expected length of the current section.

        That is computed as the difference between reference value and the expected bifurcation
        value.
        If ``bifurcation < termination``, then the reference value minus the termination will be
        computed instead.
        In real morphologies termination will be larger than bifurcation, unless ``bifurcation`` is
        set to ``inf``.
        So the expected length will be computed based on term only if the section will terminate
        before it bifurcates.
        """
        if np.isinf(self.bif):
            return abs(self.ref - self.term)
        return abs(self.ref - max(self.bif, self.term))

    def __eq__(self, other):
        if isinstance(other, TMDStop):
            return (
                self.bif_id == other.bif_id
                and self.term_id == other.term_id
                and np.isclose(self.bif, other.bif)
                and np.isclose(self.term, other.term)
                and np.allclose(self.ref, other.ref)
            )
        return False



"""Distribution configuration for the NeuroTS package."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from setuptools import find_packages
from setuptools import setup

with open("README.md", encoding="utf-8") as f:
    README = f.read()

reqs = [
    "jsonschema>=3.0.1",
    "matplotlib>=1.3.1",
    "morphio>=3.0,<4.0",
    "neurom>=3.0,<4.0",
    "numpy>=1.15.0",
    "scipy>=1.6",
    "tmd>=2.2.0",
]

doc_reqs = [
    "m2r2",
    "sphinx",
    "sphinx-bluebrain-theme",
    "sphinx-gallery",
    "sphinx-jsonschema",
]

test_reqs = [
    "diameter-synthesis>=0.5.0",
    "dictdiffer",
    "mock",
    "morph-tool>=0.1.12",
    "numpy>=1.22",
    "pytest",
    "pytest-cov",
    "pytest-html",
    "pytest-xdist",
    "tqdm",
]

setup(
    name="NeuroTS",
    author="Blue Brain Project, EPFL",
    description="Synthesis of artificial neurons using their topological profiles package",
    long_description=README,
    long_description_content_type="text/markdown",
    license="GPLv3",
    url="https://github.com/BlueBrain/NeuroTS",
    project_urls={
        "Tracker": "https://github.com/BlueBrain/NeuroTS/issues",
        "Source": "https://github.com/BlueBrain/NeuroTS",
    },
    install_requires=reqs,
    extras_require={
        "docs": doc_reqs,
        "test": test_reqs,
    },
    tests_require=test_reqs,
    python_requires=">=3.8",
    use_scm_version=True,
    setup_requires=[
        "setuptools_scm",
    ],
    packages=find_packages(include=["neurots*"]),
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
)



"""Example for comparing a neuron to a population
   using their morphometrics"""
import neurom as nm
import numpy as np
import pylab as plt

# Set a list of features to be extracted
feat_list = [
    "number_of_neurites",
    "number_of_sections_per_neurite",
    "number_of_leaves",
    "number_of_bifurcations",
    "section_lengths",
    "section_tortuosity",
    "section_radial_distances",
    "section_path_distances",
    "section_branch_orders",
    "remote_bifurcation_angles",
]

# Set a list of names for the previous features
feat_names = [
    "Number of neurites",
    "Number of sections",
    "Number of terminations",
    "Number of bifurcations",
    "Section lengths",
    "Section tortuosity",
    "Section radial distances",
    "Section path distances",
    "Section branch orders",
    "Remote bif angles",
]


def get_features(object1, object2, flist=feat_list, neurite_type=nm.BASAL_DENDRITE):
    """Computes features from module mod"""
    collect_all = []

    for feat in flist:

        feature_pop = nm.get(feat, object1, neurite_type=neurite_type)
        feature_neu = nm.get(feat, object2, neurite_type=neurite_type)

        collect_all.append([feature_pop, feature_neu])

    return collect_all


def get_features_median(object1, object2, flist=feat_list, neurite_type=nm.BASAL_DENDRITE):
    """Computes features from module mod"""
    collect_all = []

    for feat in flist:

        feature_pop = [np.median(nm.get(feat, obj, neurite_type=neurite_type)) for obj in object1]
        feature_neu = nm.get(feat, object2, neurite_type=neurite_type)

        collect_all.append([feature_pop, feature_neu])

    return collect_all


def mvs_score(data, percent=50):
    """Get the MED - MVS score equal to
    the absolute difference between the median
    of the population and the median of the neuron
    divided by the maximum visible spread.
    """
    median_diff = np.abs(np.median(data[0]) - np.median(data[1]))
    max_percentile = np.max(
        [
            np.percentile(data[0], 100 - percent / 2.0, axis=0),
            np.percentile(data[1], 100 - percent / 2.0, axis=0),
        ]
    )

    min_percentile = np.min(
        [
            np.percentile(data[0], percent / 2.0, axis=0),
            np.percentile(data[1], percent / 2.0, axis=0),
        ]
    )

    max_vis_spread = max_percentile - min_percentile

    return median_diff / max_vis_spread


def normalize(data):
    """Returns the data normalized by
    subtracting the mean of the population
    and dividing by the std of the population.
    """
    m = np.mean(data[0])
    st = np.std(data[0])

    return [(data[0] - m) / st, (data[1] - m) / st]


def score_test(data):
    """Computes the absolute differences
    between a population and a neuron and
    returns True if > 1 and False if <= 1
    """
    return np.abs(np.median(data[1]) - np.median(data[0])) > 1


def score_mvs_test(data, threshold=0.3):
    """Computes the absolute differences
    between a population and a neuron and
    returns True if > 1 and False if <= 1
    """
    return mvs_score(data) > threshold


def boxplots(data, fnames=feat_names, threshold=0.5):
    """Plots a list of boxplots for each feature in feature_list for object 1.
    Then presents the value of object 2 for each feature as an colored objected
    in the same boxplot.

    Parameters:
        data:\
            A list of pairs of flattened data for each feature.
        new_fig (Optional[bool]):\
            Default is False, which returns the default matplotlib axes 111\
            If a subplot needs to be specified, it should be provided in xxx format.
        subplot (Optional[bool]):\
            Default is False, which returns a matplotlib figure object. If True,\
            returns a matplotlib axis object, for use as a subplot.

    Returns:
        fig:\
            A figure which contains the list of boxplots.
    """
    fig = plt.figure()
    ax = fig.add_subplot(111)

    norm_data = [normalize(d) for d in data]

    ax.boxplot(np.array(norm_data)[:, 0], vert=False)

    for i, d in enumerate(norm_data):
        if score_mvs_test(data[i], threshold):
            col = "r"
            mark = "h"
        else:
            col = "g"
            mark = "s"

        ax.scatter(np.median(d[1]), len(norm_data) - i, s=100, color=col, marker=mark)

    ax.set_yticklabels(fnames)

    ax.set_xlabel("Normalized units (dimensionless)")
    ax.set_ylabel("")
    ax.set_title("Summarizing validation features")

    plt.tight_layout(True)

    return fig, ax


def dict_mvs_scores(data, feat_list):
    """Creates a dictionary with the mvs scores
    of a neuron compared to a population.
    """
    dictionary = {feat_list[i]: mvs_score(d) for i, d in enumerate(data)}

    return dictionary


def dict_mvs_scores_population(population1, population2, feat_list):
    """Creates a dictionary with the mvs scores
    of a neuron compared to a population.
    """
    dictionary = {f: [] for f in feat_list}

    for n2 in population2.neurons:
        data = get_features(population1, n2, feat_list)
        for i, d in enumerate(data):
            dictionary[feat_list[i]].append(mvs_score(d))

    return dictionary


def write_mvs_scores(dictionary, names, feat_list, output_file):
    """Outputs the dictionary of the feature scores
    for each neuron using its name as an identifier.
    """
    import csv

    F = open(output_file, "wb")
    Fcsv = csv.writer(F)
    Fcsv.writerow(["CellID"] + dictionary.keys())

    for i, n in enumerate(names):
        Fcsv.writerow([n] + list(np.array(dictionary.values())[:, i]))
    F.close()


def load_csv(filename):
    import csv

    F = open(filename)
    csv_reader = csv.reader(F)
    data = []
    for i in csv_reader:
        data.append(i)

    return np.array(np.array(data)[1:, 1:], dtype=np.float)



"""Util functions useful for general purposes."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import numpy as np
from scipy.spatial import KDTree

# The smallest representable positive number such that 1.0 + eps != 1.0
# around 1e-7 for float32
EPS = np.finfo(np.float32).eps


def get_random_point(D=1.0, random_generator=np.random):
    """Return 3-d coordinates of a new random point.

    The distance between the produced point and (0,0,0) is given by the value D.
    """
    # pylint: disable=assignment-from-no-return
    phi = random_generator.uniform(0.0, 2.0 * np.pi)
    theta = np.arccos(random_generator.uniform(-1.0, 1.0))

    sn_theta = np.sin(theta)

    x = D * np.cos(phi) * sn_theta
    y = D * np.sin(phi) * sn_theta
    z = D * np.cos(theta)

    return np.array((x, y, z))


def norm(vector):
    """Return the norm of the numpy array."""
    return np.sqrt(vector.dot(vector))


def normalize_inplace(vector):
    """Normalizes a vector in place and returns it.

    It is faster because it avoids a copy but be careful when using it because it mutates its
    argument.
    """
    vector /= norm(vector)
    return vector


def normalize_vectors(vectors):
    """Return normalized vectors."""
    return vectors / np.linalg.norm(vectors, axis=1)[:, np.newaxis]


def from_to_direction(point1, point2, return_length=False):
    """Return weight and normalized direction to target."""
    vector = point2 - point1
    length = norm(vector)

    if return_length:
        return vector / length, length

    return vector / length


def in_same_halfspace(vectors, normal, return_dots=False):
    """Return a mask of vectors that are in the same halfspace as its normal."""
    dots = vectors.dot(normal)
    mask = dots > EPS
    if return_dots:
        return mask, dots[mask]
    return mask


def in_squared_proximity(point1, point2, squared_proximity):
    """Return true if point1 is in the proximity of point2."""
    vector = point2 - point1
    return vector.dot(vector) < squared_proximity + EPS


def ball_query(points, ball_center, ball_radius):
    """Return the ids of the tree_points that are located inside the ball with center and radius."""
    tree = KDTree(
        points,
        compact_nodes=False,
        copy_data=False,
        leafsize=2**10,
        balanced_tree=False,
    )

    return np.fromiter(tree.query_ball_point(ball_center, ball_radius), dtype=int)


def upper_half_ball_query(points, ball_center, ball_radius, direction):
    """Return the points from a ball that are in the same halfspace as the given direction.

    This function returns the points that are included in the hemisphere that has ref_point as
    center, hemisphere_radius and are in the same halfspace as direction.
    """
    ids = ball_query(points, ball_center, ball_radius)

    if ids.size == 0:
        return ids

    return ids[in_same_halfspace(points[ids] - ball_center, direction)]



"""Configuration file for the Sphinx documentation builder."""

# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.

import re

from pkg_resources import get_distribution
from sphinx_gallery.sorting import ExampleTitleSortKey  # pylint: disable=import-error

# -- Project information -----------------------------------------------------

project = "NeuroTS"

# The short X.Y version
version = get_distribution(project).version

# The full version, including alpha/beta/rc tags
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx_gallery.gen_gallery",
    "sphinx-jsonschema",
    "m2r2",
]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "sphinx-bluebrain-theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

html_theme_options = {
    "metadata_distribution": project,
}

html_title = "NeuroTS"

# If true, links to the reST sources are added to the pages.
html_show_sourcelink = False

# autosummary settings
autosummary_generate = True

# autodoc settings
autodoc_typehints = "signature"
autodoc_default_options = {
    "members": True,
    "show-inheritance": True,
}

intersphinx_mapping = {
    "morphio": ("https://morphio.readthedocs.io/en/latest/", None),
    "neurom": ("https://neurom.readthedocs.io/en/stable/", None),
    "numpy": ("https://numpy.org/doc/stable/", None),
    "python": ("https://docs.python.org/3", None),
}

sphinx_gallery_conf = {
    "download_all_examples": False,
    "examples_dirs": "../../examples",
    "gallery_dirs": "examples",
    "line_numbers": True,
    "min_reported_time": 999,
    "plot_gallery": False,
    "within_subsection_order": ExampleTitleSortKey,
}


def fix_signature(app, what, name, obj, options, signature, return_annotation):
    """Remove the module locations from signatures."""
    # pylint: disable=unused-argument
    if signature:
        module_pattern = r"(.*)<module '(.*)' from '.*'>(.*)"
        match = re.match(module_pattern, signature)
        if match:
            return "".join(match.groups()), return_annotation

    return None


def setup(app):
    """Setup Sphinx by connecting functions to events."""
    app.connect("autodoc-process-signature", fix_signature)



# noqa

# Copyright (C) 2022  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Synthesize a cell with external diametrizer
===========================================

This example shows how to synthesize a cell with an external diametrizer.
An external diametrizer should have the signature described in
``neurots.generate.diametrizer.build``.

The code ``diameter_synthesis`` provides an example of such external diametrizer.

.. warning::

    Note that the ``diameter-synthesis`` package must be installed.
"""

from pathlib import Path

from diameter_synthesis import build_diameters  # pylint: disable=import-error

import neurots


def run(output_dir, data_dir):
    """Run the example for generating a cell with external diametrizer."""
    # Initialize a neuron with an external diametrizer
    N = neurots.NeuronGrower(
        input_distributions=data_dir / "IN_distr.json",
        input_parameters=data_dir / "IN_params.json",
        external_diametrizer=build_diameters.build,
    )

    # Grow the neuron
    neuron = N.grow()

    # Export the synthesized cell
    neuron.write(output_dir / "generated_cell.asc")
    neuron.write(output_dir / "generated_cell.swc")
    neuron.write(output_dir / "generated_cell.h5")


if __name__ == "__main__":
    result_dir = Path("results_neuron_external_diameter")
    result_dir.mkdir(parents=True, exist_ok=True)

    run(result_dir, Path("data"))



"""Basic class for TreeGrower Algorithms."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import logging

import numpy as np

from neurots.generate.algorithms.abstractgrower import AbstractAlgo
from neurots.generate.algorithms.common import bif_methods
from neurots.generate.algorithms.common import section_data

logger = logging.getLogger(__name__)


class TrunkAlgo(AbstractAlgo):
    """TreeGrower basic growth of trunks class.

    Args:
        input_data (dict): All the data required for the growth.
        params (dict): The parameters required for growth.
        start_point (list[float]): The first point of the trunk.
        context (Any): An object containing contextual information.
    """

    def __init__(self, input_data, params, start_point, context=None, **_):
        """Constructor of the TrunkAlgo class."""
        super().__init__(input_data, params, start_point, context)
        self.bif_method = bif_methods[params["branching_method"]]

    def initialize(self):
        """Generates the data to be used for the initialization of the first section to be grown.

        Saves the extracted input data into the corresponding structures.
        """
        stop = {"num_seg": self.params["num_seg"]}
        num_sec = 1  # A single section per tree will be generated

        return stop, num_sec

    def bifurcate(self, current_section):
        """When the section bifurcates two new sections are created.

        This method computes from the current state the data required for the
        generation of two new sections and returns the corresponding dictionaries.

        Args:
            current_section (neurots.generate.section.SectionGrowerPath): The current section.

        Returns:
            tuple[dict, dict]: Two dictionaries containing the two children sections data.
        """
        dir1, dir2 = self.bif_method()
        first_point = np.array(current_section.last_point)
        stop = current_section.stop_criteria

        return (
            section_data(dir1, first_point, stop, current_section.process),
            section_data(dir2, first_point, stop, current_section.process),
        )

    def terminate(self, current_section):
        """Terminate the current section.

        When the growth of a section is terminated the "term" must be removed from the TMD grower.
        """

    def extend(self, current_section):
        """Extend the current section.

        Create a section with the selected parameters until at least one stop criterion is
        fulfilled.
        """
        return current_section.next()


class AxonAlgo(TrunkAlgo):
    """TreeGrower of axon growth.

    Only a trunk with one segment is synthesized and another process is supposed to gaft an actual
    axon on this trunk.
    """

    def __init__(self, *args, **kwargs):
        # Force num_seg in params to 1
        params = kwargs.get("params", None) or args[1]
        params["num_seg"] = 1

        super().__init__(*args, **kwargs)



"""Input parameters functions."""

# Copyright (C) 2021  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from neurots.utils import neurite_type_warning

tmd_algos = ("tmd", "tmd_gradient", "tmd_apical")


def parameters(
    origin=(0.0, 0.0, 0.0),
    method="tmd",
    neurite_types=None,
    feature="path_distances",
    diameter_parameters=None,
):
    """Returns a default set of input parameters to be used as input for synthesis.

    Args:
        origin (list[float]): The origin point.
        method (str): The method to use.
        neurite_types (list[str]): The neurite types to consider.
        feature (str): Use the specified TMD feature.
        diameter_parameters (dict or str): The parameters used for the diameters.

    Returns:
        dict: The parameters.
    """
    if neurite_types is None:
        neurite_types = ["basal_dendrite", "apical_dendrite", "axon"]

    for i, neurite_type in enumerate(neurite_types):
        if neurite_type in ("basal", "apical"):
            neurite_type_warning(neurite_type)
            neurite_types[i] = neurite_type + "_dendrite"

    input_parameters = {
        "basal_dendrite": {},
        "apical_dendrite": {},
        "axon": {},
        "origin": list(origin),
        "grow_types": neurite_types,
    }

    def merged_params(data):
        """Use input method to set branching."""
        ret = {}
        if method == "trunk":
            branching = "random"
        elif method in tmd_algos:
            branching = "bio_oriented"
        else:
            raise KeyError(f"Method not recognized! Please select from: {tmd_algos}.")

        ret.update(
            {
                "randomness": 0.24,
                "targeting": 0.14,
                "radius": 0.3,
                "orientation": None,
                "growth_method": method,
                "branching_method": branching,
                "modify": None,
                "step_size": {"norm": {"mean": 1.0, "std": 0.2}},
                "metric": feature,
            }
        )
        ret.update(data)
        return ret

    if "axon" in neurite_types:
        input_parameters["axon"] = merged_params(
            {
                "tree_type": 2,
                "orientation": [[0.0, -1.0, 0.0]],
            }
        )

    if "basal_dendrite" in neurite_types:
        input_parameters["basal_dendrite"] = merged_params({"tree_type": 3})

    if "apical_dendrite" in neurite_types:
        input_parameters["apical_dendrite"] = merged_params(
            {
                "tree_type": 4,
                "branching_method": "directional",
                "orientation": [[0.0, 1.0, 0.0]],
            }
        )
        if method == "tmd":
            input_parameters["apical_dendrite"]["growth_method"] = "tmd_apical"

    input_parameters["diameter_params"] = {}
    if diameter_parameters is None:
        input_parameters["diameter_params"]["method"] = "default"
    elif isinstance(diameter_parameters, str):
        input_parameters["diameter_params"]["method"] = diameter_parameters
    elif isinstance(diameter_parameters, dict):
        input_parameters["diameter_params"] = diameter_parameters
        input_parameters["diameter_params"]["method"] = "external"
    else:
        raise ValueError(f"Diameter params not understood, {diameter_parameters}")

    return input_parameters



# noqa

# Copyright (C) 2022  Blue Brain Project, EPFL
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Synthesize neuron with a simple diameter model
==============================================

This example shows how to synthesize a cell with one of the simple provided diameter models.
"""

import json
from pathlib import Path

import neurots
from neurots import extract_input


def run(output_dir, data_dir):
    """Run the example for generating a cell with a simple diameter model."""
    # Extract distributions with diameters
    distr = extract_input.distributions(
        data_dir / "neurons", feature="path_distances", diameter_model="M5"
    )

    # Load default parameters dictionary
    with open(data_dir / "bio_params.json", "r", encoding="utf-8") as F:
        params = json.load(F)

    # Set the diameter method
    params["diameter_params"]["method"] = "M5"

    # Initialize a neuron
    N = neurots.NeuronGrower(input_distributions=distr, input_parameters=params)

    # Grow the neuron
    neuron = N.grow()

    # Export the synthesized cell
    neuron.write(output_dir / "generated_cell.asc")
    neuron.write(output_dir / "generated_cell.swc")
    neuron.write(output_dir / "generated_cell.h5")


if __name__ == "__main__":
    result_dir = Path("results_neuron_with_diameters")
    result_dir.mkdir(parents=True, exist_ok=True)

    run(result_dir, Path("data"))
